from fcutils.progress import track, progress
import numpy as np


def test_track():
    N = np.linspace(0, 1000, 1)

    for n in track(np.arange(N), transient=True):
        pass


def test_progress():
    N = np.linspace(0, 1000, 1)
    tid = progress.add_task("count")

    with progress:
        for n in np.arange(N):
            progress.update(tid, completed=n)
            pass



import sys

sys.path.append("./")

import numpy as np
from scipy.spatial import distance
import math

from fcutils.maths import derivative


def subtract_angles(lhs, rhs):
    """Return the signed difference between angles lhs and rhs

    Return ``(lhs - rhs)``, the value will be within ``[-math.pi, math.pi)``.
    Both ``lhs`` and ``rhs`` may either be zero-based (within
    ``[0, 2*math.pi]``), or ``-pi``-based (within ``[-math.pi, math.pi]``).
    """

    return math.fmod((lhs - rhs) + math.pi * 3, 2 * math.pi) - math.pi


def slope(x1, y1, x2, y2):
    return (y2 - y1) / (x2 - x1)


def intercept(x1, y1, x2, y2):
    return (x1 * y2 - x2 * y1) / (x1 - x2)


def get_random_point_on_line_between_two_points(x1, y1, x2, y2):
    slop = slope(x1, y1, x2, y2)
    interc = intercept(x1, y1, x2, y2)

    if np.isnan(interc):
        interc = 0

    # take a random X between the two values and compute y accordingly
    x = np.random.uniform(x1, x2)
    y = slop * x + interc
    return (x, y)


def calc_distance_between_point_and_line(line_points, p3):
    """[Calcs the perpendicular distance between a point and a line]
    
    Arguments:
        line_points {[list]} -- [list of two 2-by-1 np arrays with the two points that define the line]
        p3 {[np array]} -- [point to calculate the distance from]
    """
    p1, p2 = np.array(line_points[0]), np.array(line_points[1])
    return np.cross(p2 - p1, p3 - p1) / np.linalg.norm(p2 - p1)


def cals_distance_between_vector_and_line(line_points, v):
    dist = []
    if v.shape[1] > v.shape[0]:
        raise ValueError(
            "This function expects and NxM array with N being the number of frames and N>M, ideally M=2"
        )

    for i in range(v.shape[0]):
        p = [v[i, 0], v[i, 1]]
        dist.append(calc_distance_between_point_and_line(line_points, p))
    return dist


def calc_distance_between_points_2d(p1, p2):
    """calc_distance_between_points_2d [summary]
    
    Arguments:
        p1 {[list, array]} -- [X,Y for point one]
        p2 {[list, array]} -- [X,Y for point two]
    
    Returns:
        [float] -- [eucliden distance]

    Test: - to check : print(zero, oneh, negoneh)
    >>> zero = calc_distance_between_points_2d([0, 0], [0, 0])
    >>> oneh = calc_distance_between_points_2d([0, 0], [100, 0])
    >>> negoneh = calc_distance_between_points_2d([-100, 0], [0, 0])
    """

    return distance.euclidean(p1, p2)


def calc_distance_between_points_in_a_vector_2d(x, y):
    """
        Given a 2D array with eg X,Y tracking data it returns
        the distance between each (x,y) point
    """
    x_dot = np.abs(derivative(x))
    y_dot = np.abs(derivative(y))
    return np.sqrt(x_dot ** 2 + y_dot ** 2)


def calc_distance_between_points_two_vectors_2d(v1, v2):
    """calc_distance_between_points_two_vectors_2d [pairwise distance between vectors points]
    
    Arguments:
        v1 {[np.array]} -- [description]
        v2 {[type]} -- [description]
    
    Raises:
        ValueError -- [description]
        ValueError -- [description]
        ValueError -- [description]
    
    Returns:
        [type] -- [description]

    testing:
    >>> v1 = np.zeros((2, 5))
    >>> v2 = np.zeros((2, 5))
    >>> v2[1, :]  = [0, 10, 25, 50, 100]
    >>> d = calc_distance_between_points_two_vectors_2d(v1.T, v2.T)
    """
    # Check dataformats
    if not isinstance(v1, np.ndarray) or not isinstance(v2, np.ndarray):
        raise ValueError("Invalid argument data format")
    if not v1.shape[1] == 2 or not v2.shape[1] == 2:
        raise ValueError("Invalid shape for input arrays")
    if not v1.shape[0] == v2.shape[0]:
        raise ValueError("Error: input arrays should have the same length")

    # Calculate distance
    if v1.shape[1] < 20000 and v1.shape[0] < 20000:
        # For short vectors use cdist
        dist = distance.cdist(v1, v2, "euclidean")
        dist = dist[:, 0]
    else:
        dist = [
            calc_distance_between_points_2d(p1, p2) for p1, p2 in zip(v1, v2)
        ]
    return dist


def calc_distance_from_point(v, point):
    """[Calculates the euclidean distance from the point at each timepoint]
    
    Arguments:
        v {[np.ndarray]} -- [2D array with XY coordinates]
        point {[tuple]} -- [tuple of length 2 with X and Y coordinates of point]
    """
    assert isinstance(v, np.ndarray), "Input data needs to be a numpy array"

    if v.shape[0] == 2:
        pass  # good
    elif v.shape[1] == 2:
        v = v.T
    else:
        raise ValueError("Vector of weird shape: {}".format(v.shape))

    delta_x = v[0, :] - point[0]
    delta_y = v[1, :] - point[1]

    return np.sqrt(delta_x ** 2 + delta_y ** 2)


def calc_angles_with_arctan(x, y):
    theta = np.degrees(np.arctan2(x, y))

    if not isinstance(theta, np.ndarray):
        if theta < 0:
            theta += 360
        if theta < 0 or theta > 360:
            raise ValueError
        return theta

    theta[theta < 0] += 360
    if np.nanmax(theta) > 360 or np.nanmin(theta) < 0:
        raise ValueError("Something went wrong while computing angles")

    return theta


def calc_angle_between_points_of_vector_2d(x, y):
    """
        Given 2 1d arrays specifying for instance the X and Y coordinates at each frame,
        computes the angle between successive points (x,y)
    """
    delta_x = derivative(x)
    delta_y = derivative(y)
    return calc_angles_with_arctan(delta_x, delta_y)


def calc_angle_between_vectors_of_points_2d(x1, y1, x2, y2):
    """ 
        Given two sets of X,Y coordinates computes the angle
        between each pair of point in each set of coordinates.
    """
    # Calculate
    delta_x = np.array(x2 - x1)
    delta_y = np.array(y2 - y1)

    return calc_angles_with_arctan(delta_x, delta_y)


def calc_ang_velocity(angles):
    """calc_ang_velocity [calculates the angular velocity ]
    
    Arguments:
        angles {[np.ndarray]} -- [1d array with a timeseries of angles in degrees]
    
    Returns:
        [np.ndarray] -- [1d array with the angular velocity in degrees at each timepoint]
    
    testing:
    >>> v = calc_ang_velocity([0, 10, 100, 50, 10, 0])    
    """
    # Check input data
    if not isinstance(angles, np.ndarray) and not isinstance(angles, list):
        raise ValueError("Invalid input data format")

    if isinstance(angles, np.ndarray):
        if len(angles.shape) > 1:
            angles = angles.ravel()

    # Calculate
    angles_radis = np.unwrap(np.radians(np.nan_to_num(angles)))  # <- to unwrap
    ang_vel_rads = derivative(angles_radis)
    return np.degrees(ang_vel_rads)



from fcutils.maths import coordinates

import numpy as np
from numpy.random import uniform


def test_coordinates_transformations():
    for i in range(5):
        x, y = uniform(-100, 100), uniform(-100, 100)

        r, p = coordinates.cart2pol(x, y)

        x2, y2 = coordinates.pol2cart(r, p)

        assert np.abs(x - x2) < 0.1
        assert np.abs(y - y2) < 0.1



import pandas as pd
import numpy as np
from scipy.signal import resample


def get_onset_offset(signal, th, clean=True):
    """
        Get onset/offset times when a signal goes below>above and
        above>below a given threshold
        Arguments:
            signal: 1d numpy array
            th: float, threshold
            clean: bool. If true ends before the first start and 
                starts after the last end are removed
    """
    above = np.zeros_like(signal)
    above[signal >= th] = 1

    der = derivative(above)
    starts = np.where(der > 0)[0]
    ends = np.where(der < 0)[0]

    if above[0] > 0:
        starts = np.concatenate([[0], starts])
    if above[-1] > 0:
        ends = np.concatenate([ends, [len(signal)]])

    if clean:
        ends = np.array([e for e in ends if e > starts[0]])

        if np.any(ends):
            starts = np.array([s for s in starts if s < ends[-1]])

    if not np.any(starts):
        starts = np.array([0])
    if not np.any(ends):
        ends = np.array([len(signal)])

    return starts, ends


def smooth_hanning(x, window_len=11):
    """smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    Arguments:
        x: np.ndarray/ the input signal 
        window_len: int. the dimension of the smoothing window; should be an odd integer
    Returns:
        the smoothed signal
    
    """
    s = np.r_[x[window_len - 1 : 0 : -1], x, x[-2 : -window_len - 1 : -1]]
    w = eval("np." + window_len + "(window_len)")
    y = np.convolve(w / w.sum(), s, mode="valid")
    return y


def upsample_signal(X, original_sampling_rate, target_sampling_rate):
    """
        Updamples a signal to a target sampling rate

        Arguments:
            X: np.array with data
            original_sampling_rate: int. Sampling rate of input data
            target_sampling_rate: int. Sampling rate of output data
    """
    n_seconds = len(X) / original_sampling_rate
    goal_n_samples = np.int(n_seconds * target_sampling_rate)
    return resample(X, goal_n_samples)


def rolling_pearson_correlation(X, Y, window_size):
    """
        Computs the rolling (windowed) Pearson's
        correlation between two time series

        Arguments:
            X: 1d np.array with data
            Y: 1d np.array with data
            window_size: int. Size of rolling window
    """

    # Interpolate missing data.
    df = pd.DataFrame(dict(X=X, Y=Y))
    df_interpolated = df.interpolate()

    # Compute rolling window synchrony
    rolling_r = (
        df_interpolated["X"]
        .rolling(window=window_size, center=True)
        .corr(df_interpolated["Y"])
    )
    return rolling_r.values


def rolling_mean(X, window_size):
    """
        Computs the rolling mean of a signal

        Arguments:
            X: 1d np.array with data
            window_size: int. Size of rolling window
    """
    X = pd.Series(X)

    try:
        moving_avg = np.array(
            X.rolling(window=window_size, min_periods=1).mean(center=True)
        )
    except TypeError:  # compatible with pandas versions
        moving_avg = np.array(
            X.rolling(window=window_size, min_periods=1, center=True).mean()
        )
    return moving_avg


def derivative(X, axis=0, order=1):
    """"
        Takes the derivative of an array X along a given axis


        Arguments:
            X: np.array with data
            axis: int. Axis along which the derivative is to be computed
            order: int. Derivative order
    """

    return np.diff(X, n=order, axis=axis, prepend=0)



import numpy as np
from scipy import stats
from collections import namedtuple


def find_nearest(X, value):
    """
        Finds the index of the array value that is closest to a target value.

        Arguments:
            X: np.ndarray with data
            value: float, int. Value
    """

    idx = np.abs(X - value).argmin()
    return idx


def percentile_range(X, low=5, high=95):
    """ 
        Calculates the range between the low and high percentiles
        in an array of data X
        
        Arguments:
            X: data
            low, high: int. Low and high percentiles
    """

    lowp = np.nanpercentile(X, low)
    highp = np.nanpercentile(X, high)
    median = np.nanmedian(X)
    mean = np.nanmean(X)
    std = np.std(X)
    sem = stats.sem(X)

    res = namedtuple("percentile", "low median mean high std sem")
    return res(lowp, median, mean, highp, std, sem)



import numpy as np


from fcutils.maths.array import find_nearest, percentile_range


def plot_line_outlined(
    ax, x, y=None, lw=2, outline=1, outline_color="k", color="r", **kwargs
):
    if y is not None:
        ax.plot(x, y, lw=lw + outline, color=outline_color, zorder=-1)
        ax.plot(x, y, lw=lw, color=color, **kwargs)
    else:
        ax.plot(x, lw=lw + outline, color=outline_color, zorder=-1)
        ax.plot(x, lw=lw, color=color, **kwargs)


def plot_mean_and_error(y, yerr, ax, err_alpha=0.3, color="k", **kwargs):
    alpha = kwargs.pop("alpha", 1)
    lw = kwargs.pop("lw", 3)
    err_color = kwargs.pop("err_color", color)
    zorder = kwargs.pop("zorder", 90)
    x = kwargs.pop("x", np.arange(len(y)))

    ax.fill_between(
        x,
        y - yerr,
        y + yerr,
        alpha=err_alpha,
        zorder=zorder - 1,
        color=err_color,
    )
    ax.plot(x, y, alpha=alpha, lw=lw, zorder=zorder, color=color, **kwargs)


def ball_and_errorbar(
    x, y, data, ax, orientation="horizontal", color="k", s=100, **kwargs
):
    """
        Plots a line and a ball on top of it, used to plot
        stuff like mean and CI.
    """
    perc = percentile_range(data)
    if orientation == "horizontal":
        ax.plot([perc.low, perc.high], [y, y], color=color, **kwargs)
    else:
        ax.plot([x, x], [perc.low, perc.high], color=color, **kwargs)

    ax.scatter(x, y, color=color, s=s, zorder=99)


def ortholines(
    ax,
    orientations,
    values,
    color=[0.7, 0.7, 0.7],
    lw=3,
    alpha=0.5,
    ls="--",
    **kwargs,
):
    """[makes a set of vertical and horizzontal lines]
    
    Arguments:
        ax {[np.axarr]} -- [ax]
        orientations {[int]} -- [list of 0 and 1 with the orientation of each line. 0 = horizzontal and 1 = vertical]
        values {[float]} -- [list of x or y values at which the lines should be drawn. Should be the same length as orientations]

    """
    if not isinstance(orientations, list):
        orientations = [orientations]
    if not isinstance(values, (list, np.ndarray)):
        values = [values]

    for o, v in zip(orientations, values):
        if o == 0:
            func = ax.axhline
        else:
            func = ax.axvline

        func(v, color=color, lw=lw, alpha=alpha, ls=ls, **kwargs)


def vline_to_curve(
    ax, x, xdata, ydata, dot=False, line_kwargs={}, scatter_kwargs={}, **kwargs
):
    """[plots a vertical line from the x axis to the curve at location x]
    
    Arguments:
        ax {[axarray]} -- [ax to plot on]
        x {[float]} -- [x value to plot on ]
        curve {[np.array]} -- [array of data with the curve. The vertical line will go from 0 to curve[x]]
    """
    line = ax.plot(xdata, ydata, alpha=0)
    xline, yline = line[0].get_data()
    x = find_nearest(xline, x)
    yval = yline[np.where(xline == x)[0][0]]
    ax.plot([x, x], [0, yval], **line_kwargs)
    if dot:
        ax.scatter(x, yval, **scatter_kwargs, **kwargs)


def vline_to_point(ax, x, y, ymin=0, **kwargs):
    ax.plot([x, x], [ymin, y], **kwargs)


def hline_to_point(ax, x, y, xmin=0, **kwargs):
    ax.plot([xmin, x], [y, y], **kwargs)


def hline_to_curve(
    ax, y, xdata, ydata, dot=False, line_kwargs={}, scatter_kwargs={}, **kwargs
):
    """[plots a vertical line from the x axis to the curve at location x]
    
    Arguments:
        ax {[axarray]} -- [ax to plot on]
        x {[float]} -- [x value to plot on ]
        curve {[np.array]} -- [array of data with the curve. The vertical line will go from 0 to curve[x]]
    """
    line = ax.plot(xdata, ydata, alpha=0)
    xline, yline = line[0].get_data()
    y = find_nearest(yline, y)
    xval = xline[np.where(yline == y)[0][0]]
    ax.plot([0, xval], [y, y], **line_kwargs, **kwargs)
    if dot:
        ax.scatter(xval, y, **scatter_kwargs, **kwargs)


def plot_shaded_withline(
    ax, x, y, z=None, label=None, orientation="hor", alpha=0.15, **kwargs
):
    """[Plots a curve with shaded area and the line of the curve clearly visible]
    
    Arguments:
        ax {[type]} -- [matplotlib axis]
        x {[np.array, list]} -- [x data]
        y {[np.array, list]} -- [y data]
    
    Keyword Arguments:
        z {[type]} -- [description] (default: {None})
        label {[type]} -- [description] (default: {None})
        alpha {float} -- [description] (default: {.15})
    """

    if z == "min":  #
        z = np.nanmin(y)
    elif z == "max":
        z = np.nanmax(y)
    elif z == "mean":
        z = np.nanmean(y)

    if z is not None:
        if orientation == "hor":
            ax.fill_between(x, z, y, alpha=alpha, **kwargs)
        elif orientation == "ver":
            ax.fill_between(y, z, x, alpha=alpha, **kwargs)
        else:
            raise ValueError

    else:
        if orientation == "hor":
            ax.fill_between(x, y, alpha=alpha, **kwargs)
        elif orientation == "ver":
            ax.fill_between(y, x, alpha=alpha, **kwargs)
        else:
            raise ValueError

    ax.plot(x, y, alpha=1, label=label, **kwargs)


def rose_plot(
    ax,
    angles,
    nbins=16,
    theta_min=0,
    theta_max=360,
    density=None,
    offset=None,
    lab_unit="degrees",
    start_zero=False,
    as_hist=False,
    theta_0="N",
    **kwargs,
):
    """
    from https://stackoverflow.com/questions/22562364/circular-histogram-for-python
    Plot polar histogram of angles on ax. ax must have been created using
    subplot_kw=dict(projection='polar'). Angles are expected in radians.
    """
    if offset is None:
        ax.set_theta_offset(np.pi / 2)
    else:
        ax.set_theta_offset(offset)

    # parse kwargs
    edge_color = kwargs.pop("edge_color", "g")
    color = kwargs.pop("color", "g")
    fill = kwargs.pop("fill", False)
    linewidth = kwargs.pop("linewidth", 2)
    alpha = kwargs.pop("alpha", 1)
    xticks = kwargs.pop("xticks", True)

    # Wrap angles to [-pi, pi)
    # angles = (angles + np.pi) % (2 * np.pi) - np.pi

    # Set bins symetrically around zero
    if start_zero:
        # To have a bin edge at zero use an even number of bins
        if nbins % 2:
            nbins += 1
        bins = np.linspace(-np.pi, np.pi, num=nbins + 1)
    else:
        bins = np.linspace(
            np.radians(theta_min), np.radians(theta_max), num=nbins + 1
        )

    # Bin data and record counts
    count, bins = np.histogram(angles, bins=bins)

    # Compute width of each bins
    widths = np.diff(bins)

    # By default plot density (frequency potentially misleading)
    if density is None or density is True:
        # Area to assign each bins
        area = count / angles.size
        # Calculate corresponding bins radius
        radius = (area / np.pi) ** 0.5
    else:
        radius = count

    # Plot data on ax
    if as_hist:
        ax.bar(
            bins[:-1],
            radius,
            zorder=1,
            align="edge",
            width=widths,
            color=color,
            edgecolor=edge_color,
            fill=fill,
            linewidth=linewidth,
            alpha=alpha,
        )
    else:
        radius = np.concatenate([radius, [radius[0]]])
        if fill:
            ax.fill_between(
                bins, 0, radius, color=color, lw=linewidth, alpha=alpha
            )
        ax.plot(bins, radius, color=color, lw=linewidth, alpha=0.8)

    # Remove ylabels, they are mostly obstructive and not informative
    ax.set_yticks([])

    if not xticks:
        ax.set_xticks([])

    if lab_unit == "radians":
        label = [
            "$0$",
            r"$\pi/4$",
            r"$\pi/2$",
            r"$3\pi/4$",
            r"$\pi$",
            r"$5\pi/4$",
            r"$3\pi/2$",
            r"$7\pi/4$",
        ]
        ax.set_xticklabels(label)



from pathlib import Path
import pytest
import os
from fcutils import path


@pytest.fixture
def fld():
    return str(Path(os.getcwd()))


@pytest.fixture
def fl():
    return Path(__file__)


def test_subdirs(fld):
    from_str = path.subdirs(fld)
    from_pth = path.subdirs(Path(fld))

    assert len(from_str) == len(from_pth)


def test_files(fld):
    from_str = path.files(fld)
    from_pth = path.files(Path(fld))

    assert len(from_str) == len(from_pth)


def test_get_size(fld, fl):
    path.size(fld)

    path.size(fl)



from rich.progress import (
    Progress,
    BarColumn,
    TextColumn,
    ProgressColumn,
)
from rich.text import Text
from datetime import timedelta

from myterial import (
    orange,
    teal_light,
    light_blue_light,
)


class TimeRemainingColumn(ProgressColumn):
    """Renders estimated time remaining."""

    # Only refresh twice a second to prevent jitter
    max_refresh = 0.5

    def render(self, task):
        """Show time remaining."""
        remaining = task.time_remaining
        if remaining is None:
            return Text("-:--:--", style=teal_light)
        remaining_delta = timedelta(seconds=int(remaining))
        return Text("remaining: " + str(remaining_delta), style=teal_light)


class TimeElapsedColumn(ProgressColumn):
    """Renders estimated time elapsed."""

    # Only refresh twice a second to prevent jitter
    max_refresh = 0.5

    def render(self, task):
        """Show time elapsed."""
        elapsed = task.elapsed
        if elapsed is None:
            return Text("-:--:--", style=light_blue_light)
        elapsed_delta = timedelta(seconds=int(elapsed))
        return Text("elapsed: " + str(elapsed_delta), style=light_blue_light)


class SpeedColumn(TextColumn):
    _renderable_cache = {}

    def __init__(self, *args):
        pass

    def render(self, task):
        if task.speed is None:
            return " "
        else:
            return f"{task.speed:.1f} steps/s"


COLUMNS = [
    BarColumn(bar_width=None),
    "Completed: ",
    TextColumn("[bold magenta]Completed {task.completed}/{task.total}"),
    "[progress.percentage]{task.percentage:>3.0f}%",
    "•",
    "Speed: ",
    SpeedColumn(),
    "•",
    "Remaining: ",
    TimeRemainingColumn(),
    "Elpsed: ",
    TimeElapsedColumn(),
]

progress = Progress(*COLUMNS, transient=False,)


def track(iterable, total=None, description="Working...", transient=False):
    """
        Spawns a progress bar to monitor the progress of a for loop over
        an iterable sequence with detailed information.

        Arguments:
            iterable: list or other iterable object
            total: int. Total length of iterable
            description: str. Text to preprend to the progress bar.

        Returs:
            elements of iterable
    """
    description = f"[{orange}]" + description
    columns = [description] + COLUMNS

    if total is None:
        try:
            total = len(iterable)
        except Exception:
            raise ValueError(
                "Could not get total from iterable, pass a total value."
            )

    track_progress = Progress(*columns, transient=transient)

    with track_progress:
        yield from track_progress.track(
            iterable, total=total, description=description,
        )



from pathlib import Path
from collections import namedtuple
import math
import seaborn as sns
from matplotlib import gridspec
import matplotlib.pyplot as plt

# ---------------------------- generating figures ---------------------------- #


def create_triplot(**kwargs):
    """
        Creates a figure with one main plot and two plots on the sides
    """
    fig = plt.figure(**kwargs)
    gs = gridspec.GridSpec(2, 2, width_ratios=[4, 1], height_ratios=[4, 1])
    ax0 = plt.subplot(gs[0, 0])
    ax1 = plt.subplot(gs[0, 1])
    ax2 = plt.subplot(gs[1, 0])
    plt.tight_layout()

    axes = namedtuple("axes", "main x y")
    return fig, axes(ax0, ax2, ax1)


def calc_nrows_ncols(N, aspect=(16, 9)):
    """
        Computs the number of rows and columns to fit
        a given number N of subplots in a figure with
        aspect `aspect`.
        from: https://stackoverflow.com/questions/36482328/how-to-use-a-python-produce-as-many-subplots-of-arbitrary-size-as-necessary-acco
    """
    width = aspect[0]
    height = aspect[1]
    area = width * height * 1.0
    factor = (N / area) ** (1 / 2.0)
    cols = math.floor(width * factor)
    rows = math.floor(height * factor)
    rowFirst = width < height
    while rows * cols < N:
        if rowFirst:
            rows += 1
        else:
            cols += 1
        rowFirst = not (rowFirst)
    return rows, cols


# ---------------------------------- editing --------------------------------- #


def clean_axes(f):
    """
        Makes axes look pretty

        Arguments:
            f: matplotlib figure
    """
    ax_list = f.axes

    for ax in list(ax_list):
        try:
            sns.despine(ax=ax, offset=10, trim=False, left=False, right=True)
        except Exception:
            pass


def set_figure_subplots_aspect(
    left=0.125, right=0.9, bottom=0.06, top=0.96, wspace=0.2, hspace=0.3
):
    """
        Adjust the aspect ratio of subplots in a figure

        Arguments:
            left: the left side of the subplots of the figure
            right: the right side of the subplots of the figure
            bottom: the bottom of the subplots of the figure
            top: the top of the subplots of the figure
            wspace: the amount of width reserved for blank space between subplots
            hspace: the amount of height reserved for white space between subplots        
    """
    plt.subplots_adjust(
        left=left,  # the left side of the subplots of the figure
        right=right,  # the right side of the subplots of the figure
        bottom=bottom,  # the bottom of the subplots of the figure
        top=top,  # the top of the subplots of the figure
        wspace=wspace,  # the amount of width reserved for blank space between subplots
        hspace=hspace,  # the amount of height reserved for white space between subplots
    )


# ---------------------------------- saving ---------------------------------- #


def save_figure(f, path, svg=False, close=False):
    """
        Saves figure to file

        Arguments:
            f: matplotlib figure
            path: str, Path. Where figure will be saved
            svg: bool If true a copy of the figure will be saved as svg
            close: bool. If true the figure is closed after saving
    """
    path = Path(path).with_suffix(".png")
    f.savefig(str(path))

    if svg:
        f.savefig(str(path.with_suffix(".svg")))

    if close:
        plt.close(f)



import numpy as np
import math
from scipy import stats


# ------------------------ EXPONENTIALS AND LOGARITHMS ----------------------- #
def fexponential(x, a, b, c):
    return a * np.exp(-b * x) + c


def exponential(x, a, b, c, d):
    return a * np.exp(-c * (x - b)) + d


def flogarithmic(x, a, b, c):
    return a * np.log(b * x) + c


# --------------------------------- SIGMOIDS --------------------------------- #
def logistic(x, L, x0, k, b):
    """
    L -> shirnks the function on the Y axis. 
    x0 -> x shift. 
    k  -> slope. the smaller the flatter. Vals > 5 are good
    b -> y shift. Shifts thw whole curve up and donw
    """
    y = L / (1 + np.exp(-k * (x - x0))) + b
    return y


def centered_logistic(x, L, x0, k):
    """
    L -> shirnks the function on the Y axis. 
    x0 -> x shift. 
    k  -> slope. the smaller the flatter. Vals > 5 are good
    """
    b = (1 - L) / 2
    y = L / (1 + np.exp(-k * (x - x0))) + b
    return y


def hill_function(x, n, L, b):  # ? no work
    return L / (1 + x ** -n) + b


def hyperbolic_tangent(x, L=1, b=0, x0=0):
    return np.tanh(x - x0) / L + b


def arctangent(x, L=2, b=0, x0=0):
    return np.arctan(x - x0) / L + b


def gudermannian(x, L=2, b=0, x0=0):
    return L * np.arctan(np.tanh((x - x0) / L)) + b


def generalised_logistic(x, a, x0):  # ? no work
    """
        a > 0
    """
    if a <= 0:
        raise ValueError("Paramter 'a' should be > 0")
    return (1 + np.exp(-(x - x0))) ** -a


def algebraic_sigmoid(x, L=1, b=0, x0=0):
    return (x - x0 / (math.sqrt(1 + (x - x0) ** 2))) / L + b


def error_function(x, x0=0, scale=1, L=1, b=0):
    norm = stats.norm(x0, scale)
    return norm.cdf(x) * L + b


def centered_error_function(x, x0=0, scale=1, L=1):
    b = (1 - L) / 2
    norm = stats.norm(x0, scale)
    return norm.cdf(x) * L + b


# ------------------------------ OTHER FUNCTIONS ----------------------------- #
def linear_func(x, a, b):
    return x * a + b


def step_function(x, a, b, c):
    # Step function
    """
    a: value at x = b
    f(x) = 0 if x<b, a if x=b and 2*a if  x > a
    """
    return a * (np.sign(x - b) + c)



import json
import yaml
import shutil
import h5py

from fcutils._file import (
    raise_on_path_not_exists,
    pathify,
    return_list_smart,
    sizeof_fmt,
)

# ----------------------------------- find ----------------------------------- #
@pathify
@raise_on_path_not_exists
@return_list_smart
def subdirs(folderpath, pattern="*"):
    """
        returns all sub folders in a given folder matching a pattern
    """
    return [f for f in folderpath.glob(pattern) if f.is_dir()]


@pathify
@raise_on_path_not_exists
@return_list_smart
def files(folderpath, pattern="*"):
    """
        returns all files folders in a given folder matching a pattern
    """
    return [f for f in folderpath.glob(pattern) if f.is_file()]


# --------------------------------- deleting --------------------------------- #
@pathify
@raise_on_path_not_exists
def delete(path):  # pragma: no cover
    """
        Deletes either a folder or a file
    """
    if path.is_dir():
        shutil.rmtree(str(path))
    else:
        path.unlink()


# ----------------------------------- info ----------------------------------- #


@pathify
@raise_on_path_not_exists
def size(path, fmt=True):
    """
        Returns the size of a file or folder
        
        Arguments:
            path: str, Path to file or folder
            fmt: bool. If true a formatted string is returned
    """

    if path.is_file():
        size = path.stat().st_size
    else:
        size = sum(f.stat().st_size for f in path.glob("**/*") if f.is_file())

    if fmt:
        return sizeof_fmt(size)
    else:
        return size


# ---------------------------------- saving ---------------------------------- #
@pathify
def to_json(filepath, obj):  # pragma: no cover
    """ saves an object to json """
    if isinstance(obj, str):
        obj = json.loads(obj, indent=4, sort_keys=True)

    with open(filepath, "w") as out:
        json.dump(obj, out, indent=4, sort_keys=True)


@pathify
def to_yaml(filepath, obj):  # pragma: no cover
    """ saves an object to yaml """
    with open(filepath, "w") as out:
        yaml.dump(obj, out, default_flow_style=False, indent=4)


# ---------------------------------- loading --------------------------------- #
@pathify
@raise_on_path_not_exists
def from_json(filepath):  # pragma: no cover
    """ loads an object from json """
    with open(filepath, "r") as fin:
        return json.load(fin)


@pathify
@raise_on_path_not_exists
def from_yaml(filepath):  # pragma: no cover
    """ loads an object from yaml """
    with open(filepath, "r") as fin:
        return yaml.load(fin, Loader=yaml.FullLoader)


@pathify
@raise_on_path_not_exists
def open_hdf(filepath):  # pragma: no cover
    """
        Open and expand from a hdf file
    """
    f = h5py.File(filepath, "r")

    # List all groups and subgroups
    keys = list(f.keys())
    subkeys = {}
    for k in keys:
        try:
            subk = list(f[k].keys())
            subkeys[k] = subk
        except:
            pass

    all_keys = []
    f.visit(all_keys.append)

    return f, keys, subkeys, all_keys



import numpy as np


def R(theta):
    """
        Returns the rotation matrix necessary to remove the
        rotation of an object centered at the origin

        Arguments:
            theta: angle in degrees

        Returns:
            R: 2x2 np.ndarray with rotation matrix
    """
    theta = np.radians(theta)
    return np.array(
        [[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]]
    )


def cart2pol(x, y):
    """
        Cartesian to polar coordinates

        angles in degrees
    """
    rho = np.hypot(x, y)
    phi = np.degrees(np.arctan2(y, x))
    return rho, phi


def pol2cart(rho, phi):
    """
        Polar to cartesian coordinates

        angles in degrees
    """
    x = rho * np.cos(np.radians(phi))
    y = rho * np.sin(np.radians(phi))
    return x, y



def forceAspect(ax, aspect):
    """
        Forces the aspect ratio of an axis onto which an image what plotted with imshow.
        From:
        https://www.science-emergence.com/Articles/How-to-change-imshow-aspect-ratio-in-matplotlib-/
    """
    im = ax.get_images()
    extent = im[0].get_extent()
    ax.set_aspect(
        abs((extent[1] - extent[0]) / (extent[3] - extent[2])) / aspect
    )


def make_axes_at_center(ax):
    """
        Moves the axes splines to the center instead of the bottom left of the ax
        from https://stackoverflow.com/questions/31556446/how-to-draw-axis-in-the-middle-of-the-figure
    """
    # Move left y-axis and bottim x-axis to centre, passing through (0,0)
    ax.spines["left"].set_position("center")
    ax.spines["bottom"].set_position("center")

    # Eliminate upper and right axes
    ax.spines["right"].set_color("none")
    ax.spines["top"].set_color("none")

    # Show ticks in the left and lower axes only
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")
